__version__ = "0.0.13"
__description__ = "PolarRoute: Long-distance maritime polar route planning taking into account complex changing environmental conditions"
__license__ = "MIT"
__author__ = "Jonathan Smith, Samuel Hall, George Coombs, James Byrne,  Michael Thorne, Maria Fox"
__email__ = "jonsmi@bas.ac.uk"
__copyright__ = "2022, BAS AI Lab"
